/* ==========================================================================*/
/* - - - - - - - - - - - - -  C O D E C , C  - - - - - - - - - - - - - - - - */
/* ==========================================================================*/
/*	Created   06-14-2004   M. Wheatley                                       */
/*	Modified  09-04-2004   MW                                                */
/*.........................................................................  */
/* Module to Initialize and use the Si3000 Codec Chip.                       */
/*.........................................................................  */

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include <p30f6014.h>
#include <dci.h>
#include "common.h"
#include "timers.h"
#include "codec.h"

/*---------------------------------------------------------------------------*/
/*----------------> P R O C E S S O R    S P E C I F I C <-------------------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------> T Y P E D E F S   U N I O N S    S T R U C T U R E S <-----------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/
#define FRAMECLKRATE  (SAMPLE_RATE*256L)

#define CMD_STAGE0 0	//normal data mode
#define CMD_STAGE1 1	//need to set lsb to tell codec next word is cmd
#define CMD_STAGE2 2	//need to load output buffer with cmd word
#define CMD_STAGE3 3	//need to set for extra frame mode
#define CMD_STAGE4 4	//need to restore normal frame mode

#define QUEUE_SIZE (PROC_BLOCK_SIZE*2)	//Rx and Tx Queue size. Must be 2x
										// the process buffer size	

/*---------------------------------------------------------------------------*/
/*----------------> E X T E R N A L  R E F E R E N C E S <-------------------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------------> G L O B A L  M E M O R Y  A L L O C A T I O N <------------*/
/*---------------------------------------------------------------------------*/
unsigned int* pActiveRxBuf;
unsigned int* pActiveTxBuf;

/*---------------------------------------------------------------------------*/
/*--------------> L O C A L  M E M O R Y  A L L O C A T I O N <--------------*/
/*---------------------------------------------------------------------------*/
static volatile unsigned char CmdStage;
static volatile unsigned int CodecCmd;

static int TxQue[QUEUE_SIZE];
static int RxQue[QUEUE_SIZE];
static volatile unsigned int TxQTail;
static volatile unsigned int RxQHead;

/*---------------------------------------------------------------------------*/
/*--------------------------> P R O T O T Y P E S <--------------------------*/
/*---------------------------------------------------------------------------*/
void InitCodec(void);
void WriteCodecCtrl( unsigned char adr, unsigned char parm );
unsigned char ReadCodecCtrl( unsigned char adr );

/*---------------------------------------------------------------------------*/
/*-------------------> I N L I N E  F U N C T I O N S <----------------------*/
/*---------------------------------------------------------------------------*/

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                         -----------------                                 */
/*                        | P u t R x Q u e |                                */
/*                         -----------------                                 */
/*Places one data sample into the Receive Queue.  Called only by interrupt   */
/* service routine.  When at least half the queue is full the dataready      */
/* semaphore is set and the active pointer set to the full half of the queue */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
inline void PutRxQue(int data)
{
	RxQue[RxQHead++] = data;
	if(RxQHead==QUEUE_SIZE)	
	{
		RxQHead = 0;	//deal with buffer wrap around
		pActiveRxBuf = RxQue + QUEUE_SIZE/2;	//point to full section
		Sem.CodecRxDataRdy = TRUE;
	}
	else
	{
		if(RxQHead==QUEUE_SIZE/2)
		{
			pActiveRxBuf = RxQue;	//point to full section
			Sem.CodecRxDataRdy = TRUE;
		}
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                         -----------------                                 */
/*                        | G e t T x Q u e |                                */
/*                         -----------------                                 */
/* Gets one data sample from the Transmit Queue.  Called only by interrupt   */
/* service routine.  When at least half the queue is empty the dataready     */
/* semaphore is set and the active pointer set to the empty half of the queue*/
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
inline int GetTxQue(void)
{
static int tmp;		//make static since is inline function
	tmp = TxQue[TxQTail++]&0xFFFE;
	if(TxQTail==QUEUE_SIZE)
	{
		TxQTail = 0;	//deal with buffer wrap around
		pActiveTxBuf = TxQue+QUEUE_SIZE/2;
		Sem.CodecTxDataRdy = TRUE;
	}
	else
	{
		if(TxQTail==QUEUE_SIZE/2)
		{
			pActiveTxBuf = TxQue;
			Sem.CodecTxDataRdy = TRUE;
		}
	}
	return tmp;
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                     -------------------------                             */
/*                    | D C I I n t e r r u p t |                            */
/*                     -------------------------                             */
/*    DCI Interrupt Service routine services Si3000 CODEC data.              */
/*  Two data queues are used to hold the transmit and receive data samples   */
/*  Samples are placed and removed in this IRQ routine one sample at a time  */
/*  using the PutRxQue() and GeTXQue() routines. In the main processing loop,*/
/*  the data is removed and placed into the queue in blocks that are 1/2 the */
/*  Queue size in a sort of ping pong buffer arrangement.                    */
/*    A simple state machine inserts an extra frame into the stream so that  */
/*  commands and status can be intertwined with the data.                    */
/*    (~10uS)
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void __attribute__((__interrupt__)) _DCIInterrupt(void)
{
	switch(	CmdStage )
	{
		case CMD_STAGE0:
			TXBUF0 = GetTxQue();
 			TXBUF1 = GetTxQue();
			TXBUF2 = GetTxQue();
			TXBUF3 = GetTxQue();
			PutRxQue(RXBUF0);
			PutRxQue(RXBUF1);
			PutRxQue(RXBUF2);
			PutRxQue(RXBUF3);
			break;
		case CMD_STAGE1:
			TXBUF0 = GetTxQue();
			TXBUF1 = GetTxQue();
			TXBUF2 = GetTxQue();
			TXBUF3 = GetTxQue() | 1;	//tell codec next value is cmd
			PutRxQue(RXBUF0);
			PutRxQue(RXBUF1);
			PutRxQue(RXBUF2);
			PutRxQue(RXBUF3);
			CmdStage = CMD_STAGE2;
			break;
		case CMD_STAGE2:
			TXBUF0 = CodecCmd;  	//insert codec command
			TXBUF1 = GetTxQue();	//extra data sample
			TXBUF2 = 0x0000;		//need dummy writes
			TXBUF3 = 0x0000;		//for some reason
			PutRxQue(RXBUF0);
			PutRxQue(RXBUF1);
			PutRxQue(RXBUF2);
			PutRxQue(RXBUF3);
			CmdStage = CMD_STAGE3;
			break;
		case CMD_STAGE3:
			TXBUF0 = GetTxQue();
			TXBUF1 = GetTxQue();
			TXBUF2 = GetTxQue();
			TXBUF3 = GetTxQue();
			PutRxQue(RXBUF0);
			PutRxQue(RXBUF1);
			PutRxQue(RXBUF2);
			PutRxQue(RXBUF3);
			//set DCI to 2 buffer length
			DCICON2 = DCI_BUFF_LEN_2 & DCI_FRAME_LEN_8 & DCI_DATA_WORD_16;
			CmdStage = CMD_STAGE4;
			break;
		case CMD_STAGE4:
			TXBUF0 = GetTxQue();
			TXBUF1 = GetTxQue();
			TXBUF2 = GetTxQue();
			TXBUF3 = GetTxQue();
			CodecCmd = RXBUF0;
			PutRxQue(RXBUF1);
			//set DCI back to 4 buffer length
			DCICON2 = DCI_BUFF_LEN_4 & DCI_FRAME_LEN_16 & DCI_DATA_WORD_16;
			CmdStage = CMD_STAGE0;
			break;
	}
	IFS2bits.DCIIF = 0; /* Clear DCI interrupt flag */
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                         -------------------                               */
/*                        | I n i t C o d e c |                              */
/*                         -------------------                               */
/*    This initializes the DCI and Codec system.                             */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void InitCodec(void)
{
int i;
	for(i=0; i<QUEUE_SIZE ;i++)
	{
		RxQue[i] = 0;
		TxQue[i] = 0;
	}
	TxQTail = 0;
	RxQHead = 0;
	Sem.CodecRxDataRdy = FALSE;
	Sem.CodecTxDataRdy = FALSE;
	CmdStage = CMD_STAGE0;
	CodecCmd = 0;
	TRISFbits.TRISF6 = 1;	//Tristate Codec reset line
	/* Configure DCI receive / transmit interrupt */
	ConfigIntDCI( DCI_INT_ON & DCI_INT_PRI_6);
	/* Configure DCI module to transmit 16 bit data with multichannel mode */
	/* See dci.h for bit details */
	OpenDCI(	DCI_EN & DCI_IDLE_CON &
					DCI_DIGI_LPBACK_DIS &
					DCI_SCKD_OUP &
					DCI_FSD_OUP &
					DCI_SAMP_CLK_FAL &
					DCI_TX_ZERO_UNF &
					DCI_SDO_TRISTAT &
//					DCI_SDO_ZERO &
					DCI_DJST_OFF &
					DCI_FSM_MULTI,
				DCI_BUFF_LEN_4 & DCI_FRAME_LEN_16 & DCI_DATA_WORD_16,
				( CLOCK_FREQ / ( 2 * FRAMECLKRATE) ) - 1,
				~DCI_DIS_SLOT_0,
				~DCI_DIS_SLOT_0
			);
	LATFbits.LATF6 = 0;	//Reset Codec chip
	TRISFbits.TRISF6 = 0;
	TimeDelay(1);
	TRISFbits.TRISF6 = 1;	//Tristate Codec reset line
	TimeDelay(1);
	//send CODEC setup parameters
	WriteCodecCtrl( CODEC_N1, 1-1);		//N=1
	WriteCodecCtrl( CODEC_M1, 20-1);	//M=20
	WriteCodecCtrl( CODEC_CTRL_1,CODEC_PARM_LPD|CODEC_PARM_SPD );
	WriteCodecCtrl( CODEC_CTRL_2, 0);	//Div by 5, HPF enabled
	WriteCodecCtrl( CODEC_RXGAIN, CODEC_PARM_HIM|CODEC_PARM_MCG2|CODEC_PARM_LIG2); 
	WriteCodecCtrl( CODEC_ADCVOL, CODEC_PARM_LOM| (23<<2) ); 
	WriteCodecCtrl( CODEC_DACVOL, CODEC_PARM_SLM|CODEC_PARM_SRM|(23<<2) ); 
	WriteCodecCtrl( CODEC_ATTN, CODEC_PARM_SOT0|CODEC_PARM_LOT0); 
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      -----------------------------                        */
/*                     | W r i t e C o d e c C t r l |                       */
/*                      -----------------------------                        */
/*    Initiates a command write to the codec.                                */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void WriteCodecCtrl( unsigned char adr, unsigned char parm )
{
	CodecCmd = (unsigned int)adr;
	CodecCmd = (CodecCmd<<8) & 0x1F00;
	CodecCmd += (unsigned int)parm;
	CmdStage = CMD_STAGE1;
	while( CmdStage != CMD_STAGE0 ){}	//wait for xfer to happen
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      ----------------------------                         */
/*                     | R e a d C o d e c C t r l |                         */
/*                      ----------------------------                         */
/*    Initiates a status Read from the codec.                                */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
unsigned char ReadCodecCtrl( unsigned char adr )
{
	CodecCmd = (unsigned int)adr;
	CodecCmd = (CodecCmd<<8) & 0x1F00;
	CodecCmd |=  0x2000;
	CmdStage = CMD_STAGE1;
	while( CmdStage != CMD_STAGE0 ){}	//wait for xfer to happen
	return CodecCmd&0x00FF;
}
